@@ -22,6 +22,7 @@
 #include "rsync.h"
 #include "ifuncs.h"
 
+extern int quiet;
 extern int verbose;
 extern int dry_run;
 extern int output_motd;
@@ -31,7 +32,9 @@ extern int am_server;
 extern int am_daemon;
 extern int am_root;
 extern int rsync_port;
+extern int protect_args;
 extern int ignore_errors;
+extern int preserve_xattrs;
 extern int kluge_around_eof;
 extern int daemon_over_rsh;
 extern int sanitize_paths;
@@ -53,7 +56,7 @@ extern char *logfile_format;
 extern char *files_from;
 extern char *tmpdir;
 extern struct chmod_mode_struct *chmod_modes;
-extern struct filter_list_struct server_filter_list;
+extern struct filter_list_struct daemon_filter_list;
 extern char curr_dir[];
 #ifdef ICONV_OPTION
 extern char *iconv_opt;
@@ -211,7 +214,7 @@ int start_inband_exchange(int f_in, int f_out, const char *user, int argc, char
 	int sargc = 0;
 	char *p, *modname;
 
-	assert(argc > 0);
+	assert(argc > 0 && *argv != NULL);
 
 	if (**argv == '/') {
 		rprintf(FERROR,
@@ -313,6 +316,8 @@ int start_inband_exchange(int f_in, int f_out, const char *user, int argc, char
 
 	if (rl_nulls) {
 		for (i = 0; i < sargc; i++) {
+			if (!sargs[i]) /* stop at --protect-args NULL */
+				break;
 			write_sbuf(f_out, sargs[i]);
 			write_byte(f_out, 0);
 		}
@@ -323,6 +328,9 @@ int start_inband_exchange(int f_in, int f_out, const char *user, int argc, char
 		write_sbuf(f_out, "\n");
 	}
 
+	if (protect_args)
+		send_protected_args(f_out, sargs);
+
 	if (protocol_version < 23) {
 		if (protocol_version == 22 || !am_sender)
 			io_start_multiplex_in();
@@ -334,16 +342,24 @@ int start_inband_exchange(int f_in, int f_out, const char *user, int argc, char
 }
 
 static char *finish_pre_exec(pid_t pid, int fd, char *request,
-			     int argc, char *argv[])
+			     char **early_argv, char **argv)
 {
-	int j, status = -1;
+	int j = 0, status = -1;
 
 	if (!request)
 		request = "(NONE)";
 
 	write_buf(fd, request, strlen(request)+1);
-	for (j = 0; j < argc; j++)
+	if (early_argv) {
+		for ( ; *early_argv; early_argv++)
+			write_buf(fd, *early_argv, strlen(*early_argv)+1);
+		j = 1; /* Skip arg0 name in argv. */
+	}
+	for ( ; argv[j]; j++) {
 		write_buf(fd, argv[j], strlen(argv[j])+1);
+		if (argv[j][0] == '.' && argv[j][1] == '\0')
+			break;
+	}
 	write_byte(fd, 0);
 
 	close(fd);
@@ -383,8 +399,8 @@ static int read_arg_from_pipe(int fd, char *buf, int limit)
 
 static int rsync_module(int f_in, int f_out, int i, char *addr, char *host)
 {
-	int argc, opt_cnt;
-	char **argv, *chroot_path = NULL;
+	int argc;
+	char **argv, **orig_argv, **orig_early_argv, *chroot_path = NULL;
 	char line[BIGPATHBUFLEN];
 	uid_t uid = (uid_t)-2;  /* canonically "nobody" */
 	gid_t gid = (gid_t)-2;
@@ -489,19 +505,19 @@ static int rsync_module(int f_in, int f_out, int i, char *addr, char *host)
 		if ((p = strstr(module_dir, "/./")) != NULL) {
 			*p = '\0';
 			p += 2;
-		} else if ((p = strdup("/")) == NULL)
+		} else if ((p = strdup("/")) == NULL) /* MEMORY LEAK */
 			out_of_memory("rsync_module");
 	}
 
-	/* We do a push_dir() that doesn't actually call chdir()
+	/* We do a change_dir() that doesn't actually call chdir()
 	 * just to make a relative path absolute. */
 	strlcpy(line, curr_dir, sizeof line);
-	if (!push_dir(module_dir, 1))
+	if (!change_dir(module_dir, CD_SKIP_CHDIR))
 		goto chdir_failed;
 	if (strcmp(curr_dir, module_dir) != 0
 	 && (module_dir = strdup(curr_dir)) == NULL)
 		out_of_memory("rsync_module");
-	push_dir(line, 1); /* Restore curr_dir. */
+	change_dir(line, CD_SKIP_CHDIR); /* Restore curr_dir. */
 
 	if (use_chroot) {
 		chroot_path = module_dir;
@@ -516,25 +532,25 @@ static int rsync_module(int f_in, int f_out, int i, char *addr, char *host)
 		set_filter_dir(module_dir, module_dirlen);
 
 	p = lp_filter(i);
-	parse_rule(&server_filter_list, p, MATCHFLG_WORD_SPLIT,
-		   XFLG_ABS_IF_SLASH);
+	parse_rule(&daemon_filter_list, p, MATCHFLG_WORD_SPLIT,
+		   XFLG_ABS_IF_SLASH | XFLG_DIR2WILD3);
 
 	p = lp_include_from(i);
-	parse_filter_file(&server_filter_list, p, MATCHFLG_INCLUDE,
-	    XFLG_ABS_IF_SLASH | XFLG_OLD_PREFIXES | XFLG_FATAL_ERRORS);
+	parse_filter_file(&daemon_filter_list, p, MATCHFLG_INCLUDE,
+	    XFLG_ABS_IF_SLASH | XFLG_DIR2WILD3 | XFLG_OLD_PREFIXES | XFLG_FATAL_ERRORS);
 
 	p = lp_include(i);
-	parse_rule(&server_filter_list, p,
+	parse_rule(&daemon_filter_list, p,
 		   MATCHFLG_INCLUDE | MATCHFLG_WORD_SPLIT,
-		   XFLG_ABS_IF_SLASH | XFLG_OLD_PREFIXES);
+		   XFLG_ABS_IF_SLASH | XFLG_DIR2WILD3 | XFLG_OLD_PREFIXES);
 
 	p = lp_exclude_from(i);
-	parse_filter_file(&server_filter_list, p, 0,
-	    XFLG_ABS_IF_SLASH | XFLG_OLD_PREFIXES | XFLG_FATAL_ERRORS);
+	parse_filter_file(&daemon_filter_list, p, 0,
+	    XFLG_ABS_IF_SLASH | XFLG_DIR2WILD3 | XFLG_OLD_PREFIXES | XFLG_FATAL_ERRORS);
 
 	p = lp_exclude(i);
-	parse_rule(&server_filter_list, p, MATCHFLG_WORD_SPLIT,
-		   XFLG_ABS_IF_SLASH | XFLG_OLD_PREFIXES);
+	parse_rule(&daemon_filter_list, p, MATCHFLG_WORD_SPLIT,
+		   XFLG_ABS_IF_SLASH | XFLG_DIR2WILD3 | XFLG_OLD_PREFIXES);
 
 	log_init(1);
 
@@ -657,12 +673,12 @@ static int rsync_module(int f_in, int f_out, int i, char *addr, char *host)
 			io_printf(f_out, "@ERROR: chroot failed\n");
 			return -1;
 		}
-		if (!push_dir(module_dir, 0))
+		if (!change_dir(module_dir, CD_NORMAL))
 			goto chdir_failed;
 		if (module_dirlen)
 			sanitize_paths = 1;
 	} else {
-		if (!push_dir(module_dir, 0)) {
+		if (!change_dir(module_dir, CD_NORMAL)) {
 		  chdir_failed:
 			rsyserr(FLOG, errno, "chdir %s failed\n", module_dir);
 			io_printf(f_out, "@ERROR: chdir failed\n");
@@ -675,7 +691,7 @@ static int rsync_module(int f_in, int f_out, int i, char *addr, char *host)
 		munge_symlinks = !use_chroot || module_dirlen;
 	if (munge_symlinks) {
 		STRUCT_STAT st;
-		if (stat(SYMLINK_PREFIX, &st) == 0 && S_ISDIR(st.st_mode)) {
+		if (do_stat(SYMLINK_PREFIX, &st) == 0 && S_ISDIR(st.st_mode)) {
 			rprintf(FLOG, "Symlink munging is unsupported when a %s directory exists.\n",
 				SYMLINK_PREFIX);
 			io_printf(f_out, "@ERROR: daemon security issue -- contact admin\n", name);
@@ -730,24 +746,40 @@ static int rsync_module(int f_in, int f_out, int i, char *addr, char *host)
 
 	io_printf(f_out, "@RSYNCD: OK\n");
 
-	opt_cnt = read_args(f_in, name, line, sizeof line, rl_nulls, &argv, &argc, &request);
+	read_args(f_in, name, line, sizeof line, rl_nulls, &argv, &argc, &request);
+	orig_argv = argv;
+
+	verbose = 0; /* future verbosity is controlled by client options */
+	ret = parse_arguments(&argc, (const char ***) &argv);
+	if (protect_args && ret) {
+		orig_early_argv = orig_argv;
+		protect_args = 2;
+		read_args(f_in, name, line, sizeof line, 1, &argv, &argc, &request);
+		orig_argv = argv;
+		ret = parse_arguments(&argc, (const char ***) &argv);
+	} else
+		orig_early_argv = NULL;
 
 	if (pre_exec_pid) {
 		err_msg = finish_pre_exec(pre_exec_pid, pre_exec_fd, request,
-					  opt_cnt, argv);
+					  orig_early_argv, orig_argv);
 	}
 
-	verbose = 0; /* future verbosity is controlled by client options */
-	ret = parse_arguments(&argc, (const char ***) &argv, 0);
+	if (orig_early_argv)
+		free(orig_early_argv);
+
 	am_server = 1; /* Don't let someone try to be tricky. */
+	quiet = 0;
 	if (lp_ignore_errors(module_id))
 		ignore_errors = 1;
 	if (write_batch < 0)
 		dry_run = 1;
 
-	if (lp_fake_super(i))
+	if (lp_fake_super(i)) {
+		if (preserve_xattrs > 1)
+			preserve_xattrs = 1;
 		am_root = -1;
-	else if (am_root < 0) /* Treat --fake-super from client as --super. */
+	} else if (am_root < 0) /* Treat --fake-super from client as --super. */
 		am_root = 2;
 
 	if (filesfrom_fd == 0)
@@ -864,6 +896,17 @@ static void send_listing(int fd)
 		io_printf(fd,"@RSYNCD: EXIT\n");
 }
 
+static int load_config(int globals_only)
+{
+	if (!config_file) {
+		if (am_server && am_root <= 0)
+			config_file = RSYNCD_USERCONF;
+		else
+			config_file = RSYNCD_SYSCONF;
+	}
+	return lp_load(config_file, globals_only);
+}
+
 /* this is called when a connection is established to a client
    and we want to start talking. The setup of the system is done from
    here */
@@ -879,7 +922,7 @@ int start_daemon(int f_in, int f_out)
 	 * might cause log-file output to occur.  This ensures that the
 	 * "log file" param gets honored for the 2 non-forked use-cases
 	 * (when rsync is run by init and run by a remote shell). */
-	if (!lp_load(config_file, 0))
+	if (!load_config(0))
 		exit_cleanup(RERR_SYNTAX);
 
 	addr = client_addr(f_in);
@@ -988,13 +1031,6 @@ static void become_daemon(void)
 
 int daemon_main(void)
 {
-	if (!config_file) {
-		if (am_server && am_root <= 0)
-			config_file = RSYNCD_USERCONF;
-		else
-			config_file = RSYNCD_SYSCONF;
-	}
-
 	if (is_a_socket(STDIN_FILENO)) {
 		int i;
 
@@ -1009,7 +1045,7 @@ int daemon_main(void)
 		return start_daemon(STDIN_FILENO, STDIN_FILENO);
 	}
 
-	if (!lp_load(config_file, 1)) {
+	if (!load_config(1)) {
 		fprintf(stderr, "Failed to parse config file: %s\n", config_file);
 		exit_cleanup(RERR_SYNTAX);
 	}
