@@ -76,7 +76,7 @@ extern char curr_dir[MAXPATHLEN];
 extern struct chmod_mode_struct *chmod_modes;
 
 extern struct filter_list_struct filter_list;
-extern struct filter_list_struct server_filter_list;
+extern struct filter_list_struct daemon_filter_list;
 
 #ifdef ICONV_OPTION
 extern int filesfrom_convert;
@@ -97,7 +97,7 @@ int flist_eof = 0; /* all the file-lists are now known */
 
 #define NORMAL_NAME 0
 #define SLASH_ENDING_NAME 1
-#define DOT_NAME 2
+#define DOTDIR_NAME 2
 
 /* Starting from protocol version 26, we always use 64-bit ino_t and dev_t
  * internally, even if this platform does not allow files to have 64-bit inums.
@@ -232,33 +232,57 @@ int link_stat(const char *path, STRUCT_STAT *stp, int follow_dirlinks)
 #endif
 }
 
+static inline int is_daemon_excluded(const char *fname, int is_dir)
+{
+	if (daemon_filter_list.head
+	 && check_filter(&daemon_filter_list, FLOG, fname, is_dir) < 0) {
+		errno = ENOENT;
+		return 1;
+	}
+	return 0;
+}
+
+static inline int path_is_daemon_excluded(char *path, int ignore_filename)
+{
+	if (daemon_filter_list.head) {
+		char *slash = path;
+
+		while ((slash = strchr(slash+1, '/')) != NULL) {
+			int ret;
+			*slash = '\0';
+			ret = check_filter(&daemon_filter_list, FLOG, path, 1);
+			*slash = '/';
+			if (ret < 0) {
+				errno = ENOENT;
+				return 1;
+			}
+		}
+
+		if (!ignore_filename
+		 && check_filter(&daemon_filter_list, FLOG, path, 1) < 0) {
+			errno = ENOENT;
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
 /* This function is used to check if a file should be included/excluded
  * from the list of files based on its name and type etc.  The value of
  * filter_level is set to either SERVER_FILTERS or ALL_FILTERS. */
-static int is_excluded(char *fname, int is_dir, int filter_level)
+static int is_excluded(const char *fname, int is_dir, int filter_level)
 {
 #if 0 /* This currently never happens, so avoid a useless compare. */
 	if (filter_level == NO_FILTERS)
 		return 0;
 #endif
-	if (fname) {
-		/* never exclude '.', even if somebody does --exclude '*' */
-		if (fname[0] == '.' && !fname[1])
-			return 0;
-		/* Handle the -R version of the '.' dir. */
-		if (fname[0] == '/') {
-			int len = strlen(fname);
-			if (fname[len-1] == '.' && fname[len-2] == '/')
-				return 0;
-		}
-	}
-	if (server_filter_list.head
-	    && check_filter(&server_filter_list, fname, is_dir) < 0)
+	if (is_daemon_excluded(fname, is_dir))
 		return 1;
 	if (filter_level != ALL_FILTERS)
 		return 0;
 	if (filter_list.head
-	    && check_filter(&filter_list, fname, is_dir) < 0)
+	    && check_filter(&filter_list, FINFO, fname, is_dir) < 0)
 		return 1;
 	return 0;
 }
@@ -269,7 +293,6 @@ static void send_directory(int f, struct file_list *flist,
 static const char *pathname, *orig_dir;
 static int pathname_len;
 
-
 /* Make sure flist can hold at least flist->used + extra entries. */
 static void flist_expand(struct file_list *flist, int extra)
 {
@@ -315,34 +338,55 @@ static void flist_done_allocating(struct file_list *flist)
 		flist->pool_boundary = ptr;
 }
 
-int push_pathname(const char *dir, int len)
+/* Call this with EITHER (1) "file, NULL, 0" to chdir() to the file's
+ * F_PATHNAME(), or (2) "NULL, dir, dirlen" to chdir() to the supplied dir,
+ * with dir == NULL taken to be the starting directory, and dirlen < 0
+ * indicating that strdup(dir) should be called and then the -dirlen length
+ * value checked to ensure that it is not daemon-excluded. */
+int change_pathname(struct file_struct *file, const char *dir, int dirlen)
 {
-	if (dir == pathname)
-		return 1;
-
-	if (!orig_dir)
-		orig_dir = strdup(curr_dir);
-
-	if (pathname && !pop_dir(orig_dir)) {
-		rsyserr(FERROR, errno, "pop_dir %s failed",
-			full_fname(orig_dir));
-		exit_cleanup(RERR_FILESELECT);
-	}
-
-	if (dir && !push_dir(dir, 0)) {
-		io_error |= IOERR_GENERAL;
-		rsyserr(FERROR, errno, "push_dir %s failed in %s",
-			full_fname(dir), curr_dir);
-		return 0;
+	if (dirlen < 0) {
+		char *cpy = strdup(dir);
+		if (*cpy != '/')
+			change_dir(orig_dir, CD_SKIP_CHDIR);
+		if (path_is_daemon_excluded(cpy, 0))
+			goto chdir_error;
+		dir = cpy;
+		dirlen = -dirlen;
+	} else {
+		if (file) {
+			if (pathname == F_PATHNAME(file))
+				return 1;
+			dir = F_PATHNAME(file);
+			if (dir)
+				dirlen = strlen(dir);
+		} else if (pathname == dir)
+			return 1;
+		if (dir && *dir != '/')
+			change_dir(orig_dir, CD_SKIP_CHDIR);
 	}
 
 	pathname = dir;
-	pathname_len = len >= 0 ? len : dir ? (int)strlen(dir) : 0;
+	pathname_len = dirlen;
+
+	if (!dir)
+		dir = orig_dir;
+
+	if (!change_dir(dir, CD_NORMAL)) {
+	  chdir_error:
+		io_error |= IOERR_GENERAL;
+		rsyserr(FERROR, errno, "change_dir %s failed", full_fname(dir));
+		if (dir != orig_dir)
+			change_dir(orig_dir, CD_NORMAL);
+		pathname = NULL;
+		pathname_len = 0;
+		return 0;
+	}
 
 	return 1;
 }
 
-static void send_file_entry(int f, struct file_struct *file, int ndx, int first_ndx)
+static void send_file_entry(int f, const char *fname, struct file_struct *file, int ndx, int first_ndx)
 {
 	static time_t modtime;
 	static mode_t mode;
@@ -355,40 +399,10 @@ static void send_file_entry(int f, struct file_struct *file, int ndx, int first_
 	static gid_t gid;
 	static const char *user_name, *group_name;
 	static char lastname[MAXPATHLEN];
-	char fname[MAXPATHLEN];
 	int first_hlink_ndx = -1;
 	int l1, l2;
 	int xflags;
 
-#ifdef ICONV_OPTION
-	if (ic_send != (iconv_t)-1) {
-		xbuf outbuf, inbuf;
-
-		INIT_CONST_XBUF(outbuf, fname);
-
-		if (file->dirname) {
-			INIT_XBUF_STRLEN(inbuf, (char*)file->dirname);
-			outbuf.size -= 2; /* Reserve room for '/' & 1 more char. */
-			if (iconvbufs(ic_send, &inbuf, &outbuf, 0) < 0)
-				goto convert_error;
-			outbuf.size += 2;
-			outbuf.buf[outbuf.len++] = '/';
-		}
-
-		INIT_XBUF_STRLEN(inbuf, (char*)file->basename);
-		if (iconvbufs(ic_send, &inbuf, &outbuf, 0) < 0) {
-		  convert_error:
-			io_error |= IOERR_GENERAL;
-			rprintf(FINFO,
-			    "[%s] cannot convert filename: %s (%s)\n",
-			    who_am_i(), f_name(file, fname), strerror(errno));
-			return;
-		}
-		outbuf.buf[outbuf.len] = '\0';
-	} else
-#endif
-		f_name(file, fname);
-
 	/* Initialize starting value of xflags. */
 	if (protocol_version >= 30 && S_ISDIR(file->mode)) {
 		dir_count++;
@@ -424,28 +438,24 @@ static void send_file_entry(int f, struct file_struct *file, int ndx, int first_
 		}
 	} else if (protocol_version < 28)
 		rdev = MAKEDEV(0, 0);
-	if (preserve_uid) {
-		if ((uid_t)F_OWNER(file) == uid && *lastname)
-			xflags |= XMIT_SAME_UID;
-		else {
-			uid = F_OWNER(file);
-			if (!numeric_ids) {
-				user_name = add_uid(uid);
-				if (inc_recurse && user_name)
-					xflags |= XMIT_USER_NAME_FOLLOWS;
-			}
+	if (!preserve_uid || ((uid_t)F_OWNER(file) == uid && *lastname))
+		xflags |= XMIT_SAME_UID;
+	else {
+		uid = F_OWNER(file);
+		if (!numeric_ids) {
+			user_name = add_uid(uid);
+			if (inc_recurse && user_name)
+				xflags |= XMIT_USER_NAME_FOLLOWS;
 		}
 	}
-	if (preserve_gid) {
-		if ((gid_t)F_GROUP(file) == gid && *lastname)
-			xflags |= XMIT_SAME_GID;
-		else {
-			gid = F_GROUP(file);
-			if (!numeric_ids) {
-				group_name = add_gid(gid);
-				if (inc_recurse && group_name)
-					xflags |= XMIT_GROUP_NAME_FOLLOWS;
-			}
+	if (!preserve_gid || ((gid_t)F_GROUP(file) == gid && *lastname))
+		xflags |= XMIT_SAME_GID;
+	else {
+		gid = F_GROUP(file);
+		if (!numeric_ids) {
+			group_name = add_gid(gid);
+			if (inc_recurse && group_name)
+				xflags |= XMIT_GROUP_NAME_FOLLOWS;
 		}
 	}
 	if (file->modtime == modtime)
@@ -462,15 +472,14 @@ static void send_file_entry(int f, struct file_struct *file, int ndx, int first_
 				np->data = (void*)(long)(first_ndx + ndx + 1);
 				xflags |= XMIT_HLINK_FIRST;
 			}
-			xflags |= XMIT_HLINKED;
 		} else {
 			if (tmp_dev == dev) {
 				if (protocol_version >= 28)
 					xflags |= XMIT_SAME_DEV_pre30;
 			} else
 				dev = tmp_dev;
-			xflags |= XMIT_HLINKED;
 		}
+		xflags |= XMIT_HLINKED;
 	}
 #endif
 
@@ -675,10 +684,11 @@ static struct file_struct *recv_file_entry(struct file_list *flist,
 	}
 #endif
 
-	clean_fname(thisname, 0);
+	if (*thisname)
+		clean_fname(thisname, 0);
 
 	if (sanitize_paths)
-		sanitize_path(thisname, thisname, "", 0);
+		sanitize_path(thisname, thisname, "", 0, SP_DEFAULT);
 
 	if ((basename = strrchr(thisname, '/')) != NULL) {
 		int len = basename++ - thisname;
@@ -819,9 +829,9 @@ static struct file_struct *recv_file_entry(struct file_list *flist,
 #endif
 
 #ifdef SUPPORT_ACLS
-	/* We need one or two index int32s when we're preserving ACLs. */
-	if (preserve_acls)
-		extra_len += (S_ISDIR(mode) ? 2 : 1) * EXTRA_LEN;
+	/* Directories need an extra int32 for the default ACL. */
+	if (preserve_acls && S_ISDIR(mode))
+		extra_len += EXTRA_LEN;
 #endif
 
 	if (always_checksum && S_ISREG(mode))
@@ -854,7 +864,6 @@ static struct file_struct *recv_file_entry(struct file_list *flist,
 	bp += FILE_STRUCT_LEN;
 
 	memcpy(bp, basename, basename_len);
-	bp += basename_len + linkname_len; /* skip space for symlink too */
 
 #ifdef SUPPORT_HARD_LINKS
 	if (xflags & XMIT_HLINKED)
@@ -919,7 +928,7 @@ static struct file_struct *recv_file_entry(struct file_list *flist,
 
 #ifdef SUPPORT_LINKS
 	if (linkname_len) {
-		bp = (char*)file->basename + basename_len;
+		bp += basename_len;
 		if (first_hlink_ndx >= flist->ndx_start) {
 			struct file_struct *first = flist->files[first_hlink_ndx - flist->ndx_start];
 			memcpy(bp, F_SYMLINK(first), linkname_len);
@@ -931,7 +940,7 @@ static struct file_struct *recv_file_entry(struct file_list *flist,
 		} else {
 			read_sbuf(f, bp, linkname_len - 1);
 			if (sanitize_paths)
-				sanitize_path(bp, bp, "", lastdir_depth);
+				sanitize_path(bp, bp, "", lastdir_depth, SP_DEFAULT);
 		}
 	}
 #endif
@@ -1014,6 +1023,7 @@ struct file_struct *make_file(const char *fname, struct file_list *flist,
 	const char *basename;
 	alloc_pool_t *pool;
 	STRUCT_STAT st;
+	int excl_ret;
 	char *bp;
 
 	if (strlcpy(thisname, fname, sizeof thisname) >= sizeof thisname) {
@@ -1022,7 +1032,7 @@ struct file_struct *make_file(const char *fname, struct file_list *flist,
 	}
 	clean_fname(thisname, 0);
 	if (sanitize_paths)
-		sanitize_path(thisname, thisname, "", 0);
+		sanitize_path(thisname, thisname, "", 0, SP_DEFAULT);
 
 	if (stp && S_ISDIR(stp->st_mode)) {
 		st = *stp; /* Needed for "symlink/." with --relative. */
@@ -1096,7 +1106,17 @@ struct file_struct *make_file(const char *fname, struct file_list *flist,
 	} else
 		flags &= ~FLAG_CONTENT_DIR;
 
-	if (is_excluded(thisname, S_ISDIR(st.st_mode) != 0, filter_level)) {
+	if (S_ISDIR(st.st_mode)) {
+		if (flags & FLAG_DOTDIR_NAME) {
+			/* A "." fname (or "/." fname in relative mode) is
+			 * never excluded.  No other trailing-dotdir names
+			 * are possible. */
+			excl_ret = 0;
+		} else
+			excl_ret = is_excluded(thisname, 1, filter_level);
+	} else
+		excl_ret = is_excluded(thisname, 0, filter_level);
+	if (excl_ret) {
 		if (ignore_perishable)
 			non_perishable_cnt++;
 		return NULL;
@@ -1123,8 +1143,14 @@ struct file_struct *make_file(const char *fname, struct file_list *flist,
 			pool = dir_flist->file_pool;
 		} else
 			pool = flist->file_pool;
-	} else
+	} else {
+#ifdef SUPPORT_ACLS
+		/* Directories need an extra int32 for the default ACL. */
+		if (preserve_acls && S_ISDIR(st.st_mode))
+			extra_len += EXTRA_LEN;
+#endif
 		pool = NULL;
+	}
 
 	if (verbose > 2) {
 		rprintf(FINFO, "[%s] make_file(%s,*,%d)\n",
@@ -1172,7 +1198,6 @@ struct file_struct *make_file(const char *fname, struct file_list *flist,
 	bp += FILE_STRUCT_LEN;
 
 	memcpy(bp, basename, basename_len);
-	bp += basename_len + linkname_len; /* skip space for symlink too */
 
 #ifdef SUPPORT_HARD_LINKS
 	if (preserve_hard_links && flist && flist->prev) {
@@ -1210,16 +1235,17 @@ struct file_struct *make_file(const char *fname, struct file_list *flist,
 		file->dirname = lastdir;
 
 #ifdef SUPPORT_LINKS
-	if (linkname_len) {
-		bp = (char*)file->basename + basename_len;
-		memcpy(bp, linkname, linkname_len);
-	}
+	if (linkname_len)
+		memcpy(bp + basename_len, linkname, linkname_len);
 #endif
 
 	if (always_checksum && am_sender && S_ISREG(st.st_mode))
 		file_checksum(thisname, tmp_sum, st.st_size);
 
-	F_PATHNAME(file) = pathname;
+	if (am_sender)
+		F_PATHNAME(file) = pathname;
+	else if (!pool)
+		F_DEPTH(file) = extra_len / EXTRA_LEN;
 
 	/* This code is only used by the receiver when it is building
 	 * a list of files for a delete pass. */
@@ -1255,22 +1281,14 @@ struct file_struct *make_file(const char *fname, struct file_list *flist,
 /* Only called for temporary file_struct entries created by make_file(). */
 void unmake_file(struct file_struct *file)
 {
-	int extra_cnt = file_extra_cnt + LEN64_BUMP(file);
-#if EXTRA_ROUNDING > 0
-	if (extra_cnt & EXTRA_ROUNDING)
-		extra_cnt = (extra_cnt | EXTRA_ROUNDING) + 1;
-#endif
-	free(REQ_EXTRA(file, extra_cnt));
+	free(REQ_EXTRA(file, F_DEPTH(file)));
 }
 
 static struct file_struct *send_file_name(int f, struct file_list *flist,
-					  char *fname, STRUCT_STAT *stp,
+					  const char *fname, STRUCT_STAT *stp,
 					  int flags, int filter_level)
 {
 	struct file_struct *file;
-#if defined SUPPORT_ACLS || defined SUPPORT_XATTRS
-	stat_x sx;
-#endif
 
 	file = make_file(fname, flist, stp, flags, filter_level);
 	if (!file)
@@ -1279,28 +1297,59 @@ static struct file_struct *send_file_name(int f, struct file_list *flist,
 	if (chmod_modes && !S_ISLNK(file->mode))
 		file->mode = tweak_mode(file->mode, chmod_modes);
 
+	if (f >= 0) {
+		char fbuf[MAXPATHLEN];
+#if defined SUPPORT_ACLS || defined SUPPORT_XATTRS
+		stat_x sx;
+#endif
+
+#ifdef ICONV_OPTION
+		if (ic_send != (iconv_t)-1) {
+			xbuf outbuf, inbuf;
+
+			INIT_CONST_XBUF(outbuf, fbuf);
+
+			if (file->dirname) {
+				INIT_XBUF_STRLEN(inbuf, (char*)file->dirname);
+				outbuf.size -= 2; /* Reserve room for '/' & 1 more char. */
+				if (iconvbufs(ic_send, &inbuf, &outbuf, 0) < 0)
+					goto convert_error;
+				outbuf.size += 2;
+				outbuf.buf[outbuf.len++] = '/';
+			}
+
+			INIT_XBUF_STRLEN(inbuf, (char*)file->basename);
+			if (iconvbufs(ic_send, &inbuf, &outbuf, 0) < 0) {
+			  convert_error:
+				io_error |= IOERR_GENERAL;
+				rprintf(FINFO,
+				    "[%s] cannot convert filename: %s (%s)\n",
+				    who_am_i(), f_name(file, fbuf), strerror(errno));
+				return NULL;
+			}
+			outbuf.buf[outbuf.len] = '\0';
+		} else
+#endif
+			f_name(file, fbuf);
+
 #ifdef SUPPORT_ACLS
-	if (preserve_acls && !S_ISLNK(file->mode) && f >= 0) {
-		sx.st.st_mode = file->mode;
-		sx.acc_acl = sx.def_acl = NULL;
-		if (get_acl(fname, &sx) < 0)
-			return NULL;
-	}
+		if (preserve_acls && !S_ISLNK(file->mode)) {
+			sx.st.st_mode = file->mode;
+			sx.acc_acl = sx.def_acl = NULL;
+			if (get_acl(fname, &sx) < 0)
+				return NULL;
+		}
 #endif
 #ifdef SUPPORT_XATTRS
-	if (preserve_xattrs && f >= 0) {
-		sx.xattr = NULL;
-		if (get_xattr(fname, &sx) < 0)
-			return NULL;
-	}
+		if (preserve_xattrs) {
+			sx.xattr = NULL;
+			if (get_xattr(fname, &sx) < 0)
+				return NULL;
+		}
 #endif
 
-	maybe_emit_filelist_progress(flist->used + flist_count_offset);
+		send_file_entry(f, fbuf, file, flist->used, flist->ndx_start);
 
-	flist_expand(flist, 1);
-	flist->files[flist->used++] = file;
-	if (f >= 0) {
-		send_file_entry(f, file, flist->used - 1, flist->ndx_start);
 #ifdef SUPPORT_ACLS
 		if (preserve_acls && !S_ISLNK(file->mode)) {
 			send_acl(&sx, f);
@@ -1314,6 +1363,12 @@ static struct file_struct *send_file_name(int f, struct file_list *flist,
 		}
 #endif
 	}
+
+	maybe_emit_filelist_progress(flist->used + flist_count_offset);
+
+	flist_expand(flist, 1);
+	flist->files[flist->used++] = file;
+
 	return file;
 }
 
@@ -1521,18 +1576,16 @@ static void send_directory(int f, struct file_list *flist, char *fbuf, int len,
 	}
 }
 
-static char lastpath[MAXPATHLEN] = "";
-static int lastpath_len = 0;
-static struct file_struct *lastpath_struct;
-
 static void send_implied_dirs(int f, struct file_list *flist, char *fname,
 			      char *start, char *limit, int flags, char name_type)
 {
+	static char lastpath[MAXPATHLEN] = "";
+	static int lastpath_len = 0;
+	static struct file_struct *lastpath_struct = NULL;
 	struct file_struct *file;
 	item_list *relname_list;
 	relnamecache **rnpp;
-	char *slash;
-	int len, need_new_dir;
+	int len, need_new_dir, depth = 0;
 	struct filter_list_struct save_filter_list = filter_list;
 
 	flags = (flags | FLAG_IMPLIED_DIR) & ~(FLAG_TOP_DIR | FLAG_CONTENT_DIR);
@@ -1545,12 +1598,31 @@ static void send_implied_dirs(int f, struct file_list *flist, char *fname,
 			need_new_dir = 0;
 		else
 			need_new_dir = 1;
-	} else
+	} else {
+		char *tp = fname, *lp = lastpath;
+		/* Skip any initial directories in our path that we
+		 * have in common with lastpath. */
+		assert(start == fname);
+		for ( ; ; tp++, lp++) {
+			if (tp == limit) {
+				if (*lp == '/' || *lp == '\0')
+					goto done;
+				break;
+			}
+			if (*lp != *tp)
+				break;
+			if (*tp == '/') {
+				start = tp;
+				depth++;
+			}
+		}
 		need_new_dir = 1;
+	}
 
 	if (need_new_dir) {
 		int save_copy_links = copy_links;
 		int save_xfer_dirs = xfer_dirs;
+		char *slash;
 
 		copy_links = xfer_dirs = 1;
 
@@ -1558,7 +1630,10 @@ static void send_implied_dirs(int f, struct file_list *flist, char *fname,
 
 		for (slash = start; (slash = strchr(slash+1, '/')) != NULL; ) {
 			*slash = '\0';
-			send_file_name(f, flist, fname, NULL, flags, ALL_FILTERS);
+			file = send_file_name(f, flist, fname, NULL, flags, ALL_FILTERS);
+			depth++;
+			if (!inc_recurse && file && S_ISDIR(file->mode))
+				change_local_filter_dir(fname, strlen(fname), depth);
 			*slash = '/';
 		}
 
@@ -1567,7 +1642,8 @@ static void send_implied_dirs(int f, struct file_list *flist, char *fname,
 			if (file && !S_ISDIR(file->mode))
 				file = NULL;
 			lastpath_struct = file;
-		}
+		} else if (file && S_ISDIR(file->mode))
+			change_local_filter_dir(fname, strlen(fname), ++depth);
 
 		strlcpy(lastpath, fname, sizeof lastpath);
 		lastpath_len = limit - fname;
@@ -1611,10 +1687,8 @@ static void send1extra(int f, struct file_struct *file, struct file_list *flist)
 	f_name(file, fbuf);
 	dlen = strlen(fbuf);
 
-	if (F_PATHNAME(file) != pathname) {
-		if (!push_pathname(F_PATHNAME(file), -1))
-			exit_cleanup(RERR_FILESELECT);
-	}
+	if (!change_pathname(file, NULL, 0))
+		exit_cleanup(RERR_FILESELECT);
 
 	change_local_filter_dir(fbuf, dlen, send_dir_depth);
 
@@ -1779,13 +1853,14 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 	int64 start_write;
 	int use_ff_fd = 0;
 	int disable_buffering;
-	int flags = recurse ? FLAG_CONTENT_DIR : 0;
+	int arg_flags, flags = recurse ? FLAG_CONTENT_DIR : 0;
 	int reading_remotely = filesfrom_host != NULL;
 	int rl_flags = (reading_remotely ? 0 : RL_DUMP_COMMENTS)
 #ifdef ICONV_OPTION
 		     | (filesfrom_convert ? RL_CONVERT : 0)
 #endif
 		     | (eol_nulls || reading_remotely ? RL_EOL_NULLS : 0);
+	int implied_dot_dir = 0;
 
 	rprintf(FLOG, "building file list\n");
 	if (show_filelist_p())
@@ -1796,6 +1871,9 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 	start_write = stats.total_written;
 	gettimeofday(&start_tv, NULL);
 
+	if (!orig_dir)
+		orig_dir = strdup(curr_dir);
+
 	if (relative_paths && protocol_version >= 30)
 		implied_dirs = 1; /* We send flagged implied dirs */
 
@@ -1813,9 +1891,9 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 
 	disable_buffering = io_start_buffering_out(f);
 	if (filesfrom_fd >= 0) {
-		if (argv[0] && !push_dir(argv[0], 0)) {
-			rsyserr(FERROR_XFER, errno, "push_dir %s failed in %s",
-				full_fname(argv[0]), curr_dir);
+		if (argv[0] && !change_dir(argv[0], CD_NORMAL)) {
+			rsyserr(FERROR_XFER, errno, "change_dir %s failed",
+				full_fname(argv[0]));
 			exit_cleanup(RERR_FILESELECT);
 		}
 		use_ff_fd = 1;
@@ -1827,13 +1905,13 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 		if (use_ff_fd) {
 			if (read_line(filesfrom_fd, fbuf, sizeof fbuf, rl_flags) == 0)
 				break;
-			sanitize_path(fbuf, fbuf, "", 0);
+			sanitize_path(fbuf, fbuf, "", 0, SP_KEEP_DOT_DIRS);
 		} else {
 			if (argc-- == 0)
 				break;
 			strlcpy(fbuf, *argv++, MAXPATHLEN);
 			if (sanitize_paths)
-				sanitize_path(fbuf, fbuf, "", 0);
+				sanitize_path(fbuf, fbuf, "", 0, SP_KEEP_DOT_DIRS);
 		}
 
 		len = strlen(fbuf);
@@ -1850,7 +1928,7 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 				fbuf[len++] = '.';
 				fbuf[len] = '\0';
 			}
-			name_type = DOT_NAME;
+			name_type = DOTDIR_NAME;
 		} else if (len > 1 && fbuf[len-1] == '.' && fbuf[len-2] == '.'
 		    && (len == 2 || fbuf[len-3] == '/')) {
 			if (len + 2 >= MAXPATHLEN)
@@ -1858,9 +1936,9 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 			fbuf[len++] = '/';
 			fbuf[len++] = '.';
 			fbuf[len] = '\0';
-			name_type = DOT_NAME;
+			name_type = DOTDIR_NAME;
 		} else if (fbuf[len-1] == '.' && (len == 1 || fbuf[len-2] == '/'))
-			name_type = DOT_NAME;
+			name_type = DOTDIR_NAME;
 		else
 			name_type = NORMAL_NAME;
 
@@ -1883,8 +1961,10 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 				*p = '\0';
 				if (p == fbuf)
 					dir = "/";
-				else
+				else {
 					dir = fbuf;
+					clean_fname(dir, 0);
+				}
 				fn = p + 3;
 				while (*fn == '/')
 					fn++;
@@ -1892,20 +1972,27 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 					*--fn = '\0'; /* ensure room for '.' */
 			} else
 				fn = fbuf;
-			len = clean_fname(fn, CFN_KEEP_LEADING_DOT_DIR
-					    | CFN_KEEP_TRAILING_SLASH
+			/* A leading ./ can be used in relative mode to affect
+			 * the dest dir without its name being in the path. */
+			if (*fn == '.' && fn[1] == '/' && !implied_dot_dir) {
+				send_file_name(f, flist, ".", NULL,
+				    (flags | FLAG_IMPLIED_DIR) & ~FLAG_CONTENT_DIR,
+				    ALL_FILTERS);
+				implied_dot_dir = 1;
+			}
+			len = clean_fname(fn, CFN_KEEP_TRAILING_SLASH
 					    | CFN_DROP_TRAILING_DOT_DIR);
 			if (len == 1) {
 				if (fn[0] == '/') {
 					fn = "/.";
 					len = 2;
-					name_type = DOT_NAME;
+					name_type = DOTDIR_NAME;
 				} else if (fn[0] == '.')
-					name_type = DOT_NAME;
+					name_type = DOTDIR_NAME;
 			} else if (fn[len-1] == '/') {
 				fn[--len] = '\0';
 				if (len == 1 && *fn == '.')
-					name_type = DOT_NAME;
+					name_type = DOTDIR_NAME;
 				else
 					name_type = SLASH_ENDING_NAME;
 			}
@@ -1924,22 +2011,24 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 		if (!*fn) {
 			len = 1;
 			fn = ".";
-			name_type = DOT_NAME;
+			name_type = DOTDIR_NAME;
 		}
 
 		dirlen = dir ? strlen(dir) : 0;
 		if (dirlen != lastdir_len || memcmp(lastdir, dir, dirlen) != 0) {
-			if (!push_pathname(dir ? strdup(dir) : NULL, dirlen))
+			if (!change_pathname(NULL, dir, -dirlen))
 				continue;
 			lastdir = pathname;
 			lastdir_len = pathname_len;
-		} else if (!push_pathname(lastdir, lastdir_len))
+		} else if (!change_pathname(NULL, lastdir, lastdir_len))
 			continue;
 
 		if (fn != fbuf)
 			memmove(fbuf, fn, len + 1);
 
-		if (link_stat(fbuf, &st, copy_dirlinks || name_type != NORMAL_NAME) != 0) {
+		if (link_stat(fbuf, &st, copy_dirlinks || name_type != NORMAL_NAME) != 0
+		 || (name_type != DOTDIR_NAME && is_daemon_excluded(fbuf, S_ISDIR(st.st_mode)))
+		 || (relative_paths && path_is_daemon_excluded(fbuf, 1))) {
 			io_error |= IOERR_GENERAL;
 			rsyserr(FERROR_XFER, errno, "link_stat %s failed",
 				full_fname(fbuf));
@@ -1965,39 +2054,33 @@ struct file_list *send_file_list(int f, int argc, char *argv[])
 		} else if (implied_dirs && (p=strrchr(fbuf,'/')) && p != fbuf) {
 			/* Send the implied directories at the start of the
 			 * source spec, so we get their permissions right. */
-			char *lp = lastpath, *slash = fbuf;
-			*p = '\0';
-			/* Skip any initial directories in our path that we
-			 * have in common with lastpath. */
-			for (fn = fbuf; *fn && *lp == *fn; lp++, fn++) {
-				if (*fn == '/')
-					slash = fn;
-			}
-			*p = '/';
-			if (fn != p || (*lp && *lp != '/'))
-				send_implied_dirs(f, flist, fbuf, slash, p, flags, 0);
+			send_implied_dirs(f, flist, fbuf, fbuf, p, flags, 0);
 		}
 
 		if (one_file_system)
 			filesystem_dev = st.st_dev;
 
+		arg_flags = name_type == DOTDIR_NAME ? FLAG_DOTDIR_NAME : 0;
+
 		if (recurse || (xfer_dirs && name_type != NORMAL_NAME)) {
 			struct file_struct *file;
-			int top_flags = FLAG_TOP_DIR | FLAG_CONTENT_DIR | flags;
+			arg_flags |= FLAG_TOP_DIR | FLAG_CONTENT_DIR;
 			file = send_file_name(f, flist, fbuf, &st,
-					      top_flags, ALL_FILTERS);
+					      arg_flags | flags, ALL_FILTERS);
+			if (!file)
+				continue;
 			if (inc_recurse) {
-				if (name_type == DOT_NAME && file) {
+				if (name_type == DOTDIR_NAME) {
 					if (send_dir_depth < 0) {
 						send_dir_depth = 0;
 						change_local_filter_dir(fbuf, len, send_dir_depth);
 					}
 					send_directory(f, flist, fbuf, len, flags);
 				}
-			} else if (file)
+			} else
 				send_if_directory(f, flist, file, fbuf, len, flags);
 		} else
-			send_file_name(f, flist, fbuf, &st, flags, ALL_FILTERS);
+			send_file_name(f, flist, fbuf, &st, arg_flags | flags, ALL_FILTERS);
 	}
 
 	gettimeofday(&end_tv, NULL);
