@@ -27,7 +27,6 @@ extern int dry_run;
 extern int do_xfers;
 extern int stdout_format_has_i;
 extern int logfile_format_has_i;
-extern int receiver_symlink_times;
 extern int am_root;
 extern int am_server;
 extern int am_daemon;
@@ -98,7 +97,7 @@ extern char *backup_dir;
 extern char *backup_suffix;
 extern int backup_suffix_len;
 extern struct file_list *cur_flist, *first_flist, *dir_flist;
-extern struct filter_list_struct server_filter_list;
+extern struct filter_list_struct daemon_filter_list;
 
 int ignore_perishable = 0;
 int non_perishable_cnt = 0;
@@ -111,6 +110,7 @@ static char *deldelay_buf = NULL;
 static int deldelay_fd = -1;
 static int lull_mod;
 static int dir_tweaking;
+static int symlink_timeset_failed_flags;
 static int need_retouch_dir_times;
 static int need_retouch_dir_perms;
 static const char *solo_file = NULL;
@@ -135,8 +135,12 @@ enum delret {
     DR_SUCCESS = 0, DR_FAILURE, DR_AT_LIMIT, DR_NOT_EMPTY
 };
 
-/* Forward declaration for delete_item(). */
+/* Forward declarations. */
 static enum delret delete_dir_contents(char *fname, uint16 flags);
+#ifdef SUPPORT_HARD_LINKS
+static void handle_skipped_hlink(struct file_struct *file, int itemizing,
+				 enum logcode code, int f_out);
+#endif
 
 static int is_backup_file(char *fn)
 {
@@ -279,7 +283,7 @@ static enum delret delete_dir_contents(char *fname, uint16 flags)
 	for (j = dirlist->used; j--; ) {
 		struct file_struct *fp = dirlist->files[j];
 
-		if (fp->flags & FLAG_MOUNT_DIR) {
+		if (fp->flags & FLAG_MOUNT_DIR && S_ISDIR(fp->mode)) {
 			if (verbose > 1) {
 				rprintf(FINFO,
 				    "mount point, %s, pins parent directory\n",
@@ -511,7 +515,7 @@ static void delete_in_dir(char *fbuf, struct file_struct *file, dev_t *fs_dev)
 		struct file_struct *fp = dirlist->files[i];
 		if (!F_IS_ACTIVE(fp))
 			continue;
-		if (fp->flags & FLAG_MOUNT_DIR) {
+		if (fp->flags & FLAG_MOUNT_DIR && S_ISDIR(fp->mode)) {
 			if (verbose > 1)
 				rprintf(FINFO, "cannot delete mount point: %s\n",
 					f_name(fp, NULL));
@@ -553,10 +557,13 @@ static void do_delete_pass(void)
 	for (j = 0; j < cur_flist->used; j++) {
 		struct file_struct *file = cur_flist->sorted[j];
 
-		if (!(file->flags & FLAG_CONTENT_DIR))
-			continue;
-
 		f_name(file, fbuf);
+
+		if (!(file->flags & FLAG_CONTENT_DIR)) {
+			change_local_filter_dir(fbuf, strlen(fbuf), F_DEPTH(file));
+			continue;
+		}
+
 		if (verbose > 1 && file->flags & FLAG_TOP_DIR)
 			rprintf(FINFO, "deleting in %s\n", fbuf);
 
@@ -582,10 +589,11 @@ int unchanged_attrs(const char *fname, struct file_struct *file, stat_x *sxp)
 	if (preserve_times && cmp_time(sxp->st.st_mtime, file->modtime) != 0)
 		return 0;
 
-	if (preserve_perms && !BITS_EQUAL(sxp->st.st_mode, file->mode, CHMOD_BITS))
-		return 0;
-
-	if (preserve_executability && ((sxp->st.st_mode & 0111 ? 1 : 0) ^ (file->mode & 0111 ? 1 : 0)))
+	if (preserve_perms) {
+		if (!BITS_EQUAL(sxp->st.st_mode, file->mode, CHMOD_BITS))
+			return 0;
+	} else if (preserve_executability
+	 && ((sxp->st.st_mode & 0111 ? 1 : 0) ^ (file->mode & 0111 ? 1 : 0)))
 		return 0;
 
 	if (am_root && uid_ndx && sxp->st.st_uid != (uid_t)F_OWNER(file))
@@ -622,24 +630,31 @@ void itemize(const char *fnamecmp, struct file_struct *file, int ndx, int statre
 		int keep_time = !preserve_times ? 0
 		    : S_ISDIR(file->mode) ? preserve_times > 1 :
 #if defined HAVE_LUTIMES && defined HAVE_UTIMES
-		    (receiver_symlink_times && !(file->flags & FLAG_TIME_FAILED)) ||
-#endif
+		    1;
+#else
 		    !S_ISLNK(file->mode);
+#endif
 
 		if (S_ISREG(file->mode) && F_LENGTH(file) != sxp->st.st_size)
 			iflags |= ITEM_REPORT_SIZE;
-		if ((iflags & (ITEM_TRANSFER|ITEM_LOCAL_CHANGE) && !keep_time
-		  && !(iflags & ITEM_MATCHED)
+		if (file->flags & FLAG_TIME_FAILED) { /* symlinks only */
+			if (iflags & ITEM_LOCAL_CHANGE)
+				iflags |= symlink_timeset_failed_flags;
+		} else if (keep_time
+		 ? cmp_time(file->modtime, sxp->st.st_mtime) != 0
+		 : iflags & (ITEM_TRANSFER|ITEM_LOCAL_CHANGE) && !(iflags & ITEM_MATCHED)
 		  && (!(iflags & ITEM_XNAME_FOLLOWS) || *xname))
-		 || (keep_time && cmp_time(file->modtime, sxp->st.st_mtime) != 0))
 			iflags |= ITEM_REPORT_TIME;
 #if !defined HAVE_LCHMOD && !defined HAVE_SETATTRLIST
 		if (S_ISLNK(file->mode)) {
 			;
 		} else
 #endif
-		if ((preserve_perms || preserve_executability)
-		 && !BITS_EQUAL(sxp->st.st_mode, file->mode, CHMOD_BITS))
+		if (preserve_perms) {
+			if (!BITS_EQUAL(sxp->st.st_mode, file->mode, CHMOD_BITS))
+				iflags |= ITEM_REPORT_PERMS;
+		} else if (preserve_executability
+		 && ((sxp->st.st_mode & 0111 ? 1 : 0) ^ (file->mode & 0111 ? 1 : 0)))
 			iflags |= ITEM_REPORT_PERMS;
 		if (uid_ndx && am_root && (uid_t)F_OWNER(file) != sxp->st.st_uid)
 			iflags |= ITEM_REPORT_OWNER;
@@ -1155,8 +1170,7 @@ static int try_dests_non(struct file_struct *file, char *fname, int ndx,
 		if (itemizing && stdout_format_has_i
 		 && (verbose > 1 || stdout_format_has_i > 1)) {
 			int chg = compare_dest && type != TYPE_DIR ? 0
-			    : ITEM_LOCAL_CHANGE
-			     + (match_level == 3 ? ITEM_XNAME_FOLLOWS : 0);
+			    : ITEM_LOCAL_CHANGE + (match_level == 3 ? ITEM_XNAME_FOLLOWS : 0);
 			char *lp = match_level == 3 ? "" : NULL;
 			itemize(cmpbuf, file, ndx, 0, sxp, chg + ITEM_MATCHED, 0, lp);
 		}
@@ -1202,22 +1216,34 @@ static void list_file_entry(struct file_struct *f)
 static int phase = 0;
 static int dflt_perms;
 
+static int implied_dirs_are_missing;
+/* Helper for recv_generator's skip_dir and dry_missing_dir tests. */
+static BOOL is_below(struct file_struct *file, struct file_struct *subtree)
+{
+	return F_DEPTH(file) > F_DEPTH(subtree)
+		&& (!implied_dirs_are_missing || f_name_has_prefix(file, subtree));
+}
+
 /* Acts on the indicated item in cur_flist whose name is fname.  If a dir,
  * make sure it exists, and has the right permissions/timestamp info.  For
  * all other non-regular files (symlinks, etc.) we create them here.  For
  * regular files that have changed, we try to find a basis file and then
  * start sending checksums.  The ndx is the file's unique index value.
  *
- * When fname is non-null, it must point to a MAXPATHLEN buffer!
+ * The fname parameter must point to a MAXPATHLEN buffer!  (e.g it gets
+ * passed to delete_item(), which can use it during a recursive delete.)
  *
  * Note that f_out is set to -1 when doing final directory-permission and
  * modification-time repair. */
 static void recv_generator(char *fname, struct file_struct *file, int ndx,
 			   int itemizing, enum logcode code, int f_out)
 {
-	static int missing_below = -1, excluded_below = -1;
 	static const char *parent_dirname = "";
-	static struct file_struct *missing_dir = NULL, *excluded_dir = NULL;
+	/* Missing dir not created due to --dry-run; will still be scanned. */
+	static struct file_struct *dry_missing_dir = NULL;
+	/* Missing dir whose contents are skipped altogether due to
+	 * --ignore-non-existing, daemon exclude, or mkdir failure. */
+	static struct file_struct *skip_dir = NULL;
 	static struct file_list *fuzzy_dirlist = NULL;
 	static int need_fuzzy_dirlist = 0;
 	struct file_struct *fuzzy_file = NULL;
@@ -1229,7 +1255,6 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 	char *fnamecmp, *partialptr, *backupptr = NULL;
 	char fnamecmpbuf[MAXPATHLEN];
 	uchar fnamecmp_type;
-	int implied_dirs_are_missing = relative_paths && !implied_dirs && protocol_version < 30;
 	int del_opts = delete_mode || force_delete ? DEL_RECURSE : 0;
 	int is_dir = !S_ISDIR(file->mode) ? 0
 		   : inc_recurse && ndx != cur_flist->ndx_start - 1 ? -1
@@ -1246,48 +1271,44 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 		return;
 	}
 
-	if (server_filter_list.head) {
-		int filtered = check_filter(&server_filter_list, fname, is_dir) < 0;
-		if (is_dir < 0 && filtered)
+	if (skip_dir) {
+		if (is_below(file, skip_dir)) {
+			if (is_dir)
+				file->flags |= FLAG_MISSING_DIR;
+#ifdef SUPPORT_HARD_LINKS
+			else if (F_IS_HLINKED(file))
+				handle_skipped_hlink(file, itemizing, code, f_out);
+#endif
 			return;
-		if (excluded_below >= 0) {
-			if (F_DEPTH(file) > excluded_below
-			 && (!implied_dirs_are_missing || f_name_has_prefix(file, excluded_dir)))
-				goto skipping;
-			excluded_below = -1;
 		}
-		if (filtered) {
-			if (is_dir) {
-				excluded_below = F_DEPTH(file);
-				excluded_dir = file;
-			}
-		  skipping:
+		skip_dir = NULL;
+	}
+
+	if (daemon_filter_list.head) {
+		if (check_filter(&daemon_filter_list, FLOG, fname, is_dir) < 0) {
+			if (is_dir < 0)
+				return;
+#ifdef SUPPORT_HARD_LINKS
+			if (F_IS_HLINKED(file))
+				handle_skipped_hlink(file, itemizing, code, f_out);
+#endif
 			rprintf(FERROR_XFER,
-				"skipping daemon-excluded file \"%s\"\n",
-				fname);
+				"skipping daemon-excluded %s \"%s\"\n",
+				is_dir ? "directory" : "file", fname);
+			if (is_dir)
+				goto skipping_dir_contents;
 			return;
 		}
 	}
 
-	if (missing_below >= 0) {
-		if (F_DEPTH(file) <= missing_below
-		 || (implied_dirs_are_missing && !f_name_has_prefix(file, missing_dir))) {
-			if (dry_run)
-				dry_run--;
-			missing_below = -1;
-		} else if (!dry_run) {
-			if (is_dir)
-				file->flags |= FLAG_MISSING_DIR;
-			return;
-		}
-	}
 #ifdef SUPPORT_ACLS
 	sx.acc_acl = sx.def_acl = NULL;
 #endif
 #ifdef SUPPORT_XATTRS
 	sx.xattr = NULL;
 #endif
-	if (dry_run > 1) {
+	if (dry_run > 1 || (dry_missing_dir && is_below(file, dry_missing_dir))) {
+	  parent_is_dry_missing:
 		if (fuzzy_dirlist) {
 			flist_free(fuzzy_dirlist);
 			fuzzy_dirlist = NULL;
@@ -1297,13 +1318,17 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 		stat_errno = ENOENT;
 	} else {
 		const char *dn = file->dirname ? file->dirname : ".";
+		dry_missing_dir = NULL;
 		if (parent_dirname != dn && strcmp(parent_dirname, dn) != 0) {
 			if (relative_paths && !implied_dirs
-			 && do_stat(dn, &sx.st) < 0
-			 && create_directory_path(fname) < 0) {
-				rsyserr(FERROR_XFER, errno,
-					"recv_generator: mkdir %s failed",
-					full_fname(dn));
+			 && do_stat(dn, &sx.st) < 0) {
+				if (dry_run)
+					goto parent_is_dry_missing;
+				if (create_directory_path(fname) < 0) {
+					rsyserr(FERROR_XFER, errno,
+						"recv_generator: mkdir %s failed",
+						full_fname(dn));
+				}
 			}
 			if (fuzzy_dirlist) {
 				flist_free(fuzzy_dirlist);
@@ -1332,14 +1357,13 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 		if (is_dir) {
 			if (is_dir < 0)
 				return;
-			if (missing_below < 0) {
-				if (dry_run)
-					dry_run++;
-				missing_below = F_DEPTH(file);
-				missing_dir = file;
-			}
+			skip_dir = file;
 			file->flags |= FLAG_MISSING_DIR;
 		}
+#ifdef SUPPORT_HARD_LINKS
+		else if (F_IS_HLINKED(file))
+			handle_skipped_hlink(file, itemizing, code, f_out);
+#endif
 		if (verbose > 1) {
 			rprintf(FINFO, "not creating new %s \"%s\"\n",
 				is_dir ? "directory" : "file", fname);
@@ -1351,6 +1375,17 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 	 && !am_root && sx.st.st_uid == our_uid)
 		del_opts |= DEL_NO_UID_WRITE;
 
+	if (ignore_existing > 0 && statret == 0
+	 && (!is_dir || !S_ISDIR(sx.st.st_mode))) {
+		if (verbose > 1 && is_dir >= 0)
+			rprintf(FINFO, "%s exists\n", fname);
+#ifdef SUPPORT_HARD_LINKS
+		if (F_IS_HLINKED(file))
+			handle_skipped_hlink(file, itemizing, code, f_out);
+#endif
+		goto cleanup;
+	}
+
 	if (is_dir) {
 		if (!implied_dirs && file->flags & FLAG_IMPLIED_DIR)
 			goto cleanup;
@@ -1378,10 +1413,10 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 				goto skipping_dir_contents;
 			statret = -1;
 		}
-		if (dry_run && statret != 0 && missing_below < 0) {
-			missing_below = F_DEPTH(file);
-			missing_dir = file;
-			dry_run++;
+		if (dry_run && statret != 0) {
+			if (!dry_missing_dir)
+				dry_missing_dir = file;
+			file->flags |= FLAG_MISSING_DIR;
 		}
 		real_ret = statret;
 		real_sx = sx;
@@ -1414,8 +1449,7 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 			  skipping_dir_contents:
 				rprintf(FERROR,
 				    "*** Skipping any contents from this failed directory ***\n");
-				missing_below = F_DEPTH(file);
-				missing_dir = file;
+				skip_dir = file;
 				file->flags |= FLAG_MISSING_DIR;
 				goto cleanup;
 			}
@@ -1448,9 +1482,13 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 				DEV_MINOR(devp) = minor(real_sx.st.st_dev);
 			}
 		}
-		else if (delete_during && f_out != -1 && !phase && dry_run < 2
-		    && (file->flags & FLAG_CONTENT_DIR))
-			delete_in_dir(fname, file, &real_sx.st.st_dev);
+		else if (delete_during && f_out != -1 && !phase
+		    && !(file->flags & FLAG_MISSING_DIR)) {
+			if (file->flags & FLAG_CONTENT_DIR)
+				delete_in_dir(fname, file, &real_sx.st.st_dev);
+			else
+				change_local_filter_dir(fname, strlen(fname), F_DEPTH(file));
+		}
 		goto cleanup;
 	}
 
@@ -1534,7 +1572,7 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 			set_file_attrs(fname, file, NULL, NULL, 0);
 			if (itemizing) {
 				itemize(fname, file, ndx, statret, &sx,
-					ITEM_LOCAL_CHANGE, 0, NULL);
+					ITEM_LOCAL_CHANGE|ITEM_REPORT_CHANGE, 0, NULL);
 			}
 			if (code != FNONE && verbose)
 				rprintf(code, "%s -> %s\n", fname, sl);
@@ -1618,7 +1656,7 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 			set_file_attrs(fname, file, NULL, NULL, 0);
 			if (itemizing) {
 				itemize(fname, file, ndx, statret, &sx,
-					ITEM_LOCAL_CHANGE, 0, NULL);
+					ITEM_LOCAL_CHANGE|ITEM_REPORT_CHANGE, 0, NULL);
 			}
 			if (code != FNONE && verbose)
 				rprintf(code, "%s\n", fname);
@@ -1656,16 +1694,14 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 		goto cleanup;
 	}
 
-	if (ignore_existing > 0 && statret == 0) {
-		if (verbose > 1)
-			rprintf(FINFO, "%s exists\n", fname);
-		goto cleanup;
-	}
-
 	if (update_only > 0 && statret == 0
 	    && cmp_time(sx.st.st_mtime, file->modtime) > 0) {
 		if (verbose > 1)
 			rprintf(FINFO, "%s is newer\n", fname);
+#ifdef SUPPORT_HARD_LINKS
+		if (F_IS_HLINKED(file))
+			handle_skipped_hlink(file, itemizing, code, f_out);
+#endif
 		goto cleanup;
 	}
 
@@ -1705,7 +1741,7 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 	} else
 		partialptr = NULL;
 
-	if (statret != 0 && fuzzy_dirlist && dry_run <= 1) {
+	if (statret != 0 && fuzzy_dirlist) {
 		int j = find_fuzzy(file, fuzzy_dirlist);
 		if (j >= 0) {
 			fuzzy_file = fuzzy_dirlist->files[j];
@@ -1735,9 +1771,6 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 		goto cleanup;
 	}
 
-	if (append_mode > 0 && sx.st.st_size >= F_LENGTH(file))
-		goto cleanup;
-
 	if (fnamecmp_type <= FNAMECMP_BASIS_DIR_HIGH)
 		;
 	else if (fnamecmp_type == FNAMECMP_FUZZY)
@@ -1762,6 +1795,14 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 		goto cleanup;
 	}
 
+	if (append_mode > 0 && sx.st.st_size >= F_LENGTH(file)) {
+#ifdef SUPPORT_HARD_LINKS
+		if (F_IS_HLINKED(file))
+			handle_skipped_hlink(file, itemizing, code, f_out);
+#endif
+		goto cleanup;
+	}
+
   prepare_to_open:
 	if (partialptr) {
 		sx.st = partial_st;
@@ -1849,7 +1890,7 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 		rprintf(FINFO, "generating and sending sums for %d\n", ndx);
 
   notify_others:
-	if (remove_source_files && !delay_updates && !phase)
+	if (remove_source_files && !delay_updates && !phase && !dry_run)
 		increment_active_files(ndx, itemizing, code);
 	if (inc_recurse && !dry_run)
 		cur_flist->in_progress++;
@@ -1861,7 +1902,7 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 	if (itemizing) {
 		int iflags = ITEM_TRANSFER;
 		if (always_checksum > 0)
-			iflags |= ITEM_REPORT_CHECKSUM;
+			iflags |= ITEM_REPORT_CHANGE;
 		if (fnamecmp_type != FNAMECMP_FNAME)
 			iflags |= ITEM_BASIS_TYPE_FOLLOWS;
 		if (fnamecmp_type == FNAMECMP_FUZZY)
@@ -1918,6 +1959,28 @@ static void recv_generator(char *fname, struct file_struct *file, int ndx,
 	return;
 }
 
+#ifdef SUPPORT_HARD_LINKS
+static void handle_skipped_hlink(struct file_struct *file, int itemizing,
+				 enum logcode code, int f_out)
+{
+	char fbuf[MAXPATHLEN];
+	int new_last_ndx;
+	struct file_list *save_flist = cur_flist;
+
+	/* If we skip the last item in a chain of links and there was a
+	 * prior non-skipped hard-link waiting to finish, finish it now. */
+	if ((new_last_ndx = skip_hard_link(file, &cur_flist)) < 0)
+		return;
+
+	file = cur_flist->files[new_last_ndx - cur_flist->ndx_start];
+	cur_flist->in_progress--; /* undo prior increment */
+	f_name(file, fbuf);
+	recv_generator(fbuf, file, new_last_ndx, itemizing, code, f_out);
+
+	cur_flist = save_flist;
+}
+#endif
+
 static void touch_up_dirs(struct file_list *flist, int ndx)
 {
 	static int counter = 0;
@@ -1949,8 +2012,12 @@ static void touch_up_dirs(struct file_list *flist, int ndx)
 		fname = f_name(file, NULL);
 		if (!(file->mode & S_IWUSR))
 			do_chmod(fname, file->mode);
-		if (need_retouch_dir_times)
-			set_modtime(fname, file->modtime, file->mode);
+		if (need_retouch_dir_times) {
+			STRUCT_STAT st;
+			if (link_stat(fname, &st, 0) == 0
+			 && cmp_time(st.st_mtime, file->modtime) != 0)
+				set_modtime(fname, file->modtime, file->mode);
+		}
 		if (allowed_lull && !(counter % lull_mod))
 			maybe_send_keepalive();
 		else if (!(counter & 0xFF))
@@ -2068,6 +2135,9 @@ void generate_files(int f_out, const char *local_name)
 	dir_tweaking = !(list_only || solo_file || dry_run);
 	need_retouch_dir_times = preserve_times > 1;
 	lull_mod = allowed_lull * 5;
+	symlink_timeset_failed_flags = ITEM_REPORT_TIME
+	    | (protocol_version >= 30 || !am_server ? ITEM_REPORT_TIMEFAIL : 0);
+	implied_dirs_are_missing = relative_paths && !implied_dirs && protocol_version < 30;
 
 	if (verbose > 2)
 		rprintf(FINFO, "generator starting pid=%ld\n", (long)getpid());
@@ -2112,16 +2182,18 @@ void generate_files(int f_out, const char *local_name)
 			f_name(fp, fbuf);
 			ndx = cur_flist->ndx_start - 1;
 			recv_generator(fbuf, fp, ndx, itemizing, code, f_out);
-			if (delete_during && dry_run < 2 && !list_only) {
-				if (BITS_SETnUNSET(fp->flags, FLAG_CONTENT_DIR, FLAG_MISSING_DIR)) {
+			if (delete_during && dry_run < 2 && !list_only
+			 && !(fp->flags & FLAG_MISSING_DIR)) {
+				if (fp->flags & FLAG_CONTENT_DIR) {
 					dev_t dirdev;
 					if (one_file_system) {
 						uint32 *devp = F_DIR_DEV_P(fp);
 						dirdev = MAKEDEV(DEV_MAJOR(devp), DEV_MINOR(devp));
 					} else
 						dirdev = MAKEDEV(0, 0);
-					delete_in_dir(f_name(fp, fbuf), fp, &dirdev);
-				}
+					delete_in_dir(fbuf, fp, &dirdev);
+				} else
+					change_local_filter_dir(fbuf, strlen(fbuf), F_DEPTH(fp));
 			}
 		}
 		for (i = cur_flist->low; i <= cur_flist->high; i++) {
