@@ -78,10 +78,7 @@ extern char *batch_name;
 extern char *password_file;
 extern char curr_dir[MAXPATHLEN];
 extern struct file_list *first_flist;
-extern struct filter_list_struct server_filter_list;
-#ifdef ICONV_OPTION
-extern iconv_t ic_send;
-#endif
+extern struct filter_list_struct daemon_filter_list;
 
 uid_t our_uid;
 int local_server = 0;
@@ -334,7 +331,7 @@ static pid_t do_cmd(char *cmd, char *machine, char *user, char **remote_argv, in
 {
 	int i, argc = 0;
 	char *args[MAX_ARGS];
-	pid_t ret;
+	pid_t pid;
 	int dash_l_set = 0;
 
 	if (!read_batch && !local_server) {
@@ -344,7 +341,7 @@ static pid_t do_cmd(char *cmd, char *machine, char *user, char **remote_argv, in
 			cmd = rsh_env;
 		if (!cmd)
 			cmd = RSYNC_RSH;
-		cmd = strdup(cmd);
+		cmd = strdup(cmd); /* MEMORY LEAK */
 		if (!cmd)
 			goto oom;
 
@@ -453,7 +450,7 @@ static pid_t do_cmd(char *cmd, char *machine, char *user, char **remote_argv, in
 		batch_gen_fd = from_gen_pipe[0];
 		*f_out_p = from_gen_pipe[1];
 		*f_in_p = batch_fd;
-		ret = -1; /* no child pid */
+		pid = (pid_t)-1; /* no child pid */
 #ifdef ICONV_CONST
 		setup_iconv();
 #endif
@@ -463,54 +460,20 @@ static pid_t do_cmd(char *cmd, char *machine, char *user, char **remote_argv, in
 		if (whole_file < 0 && !write_batch)
 			whole_file = 1;
 		set_allow_inc_recurse();
-		ret = local_child(argc, args, f_in_p, f_out_p, child_main);
+		pid = local_child(argc, args, f_in_p, f_out_p, child_main);
 #ifdef ICONV_CONST
 		setup_iconv();
 #endif
 	} else {
+		pid = piped_child(args, f_in_p, f_out_p);
 #ifdef ICONV_CONST
 		setup_iconv();
 #endif
-		if (protect_args) {
-			int fd;
-#ifdef ICONV_OPTION
-			int convert = ic_send != (iconv_t)-1;
-			xbuf outbuf, inbuf;
-
-			if (convert)
-				alloc_xbuf(&outbuf, 1024);
-#endif
-
-			ret = piped_child(args, f_in_p, f_out_p);
-
-			for (i = 0; args[i]; i++) {} /* find first NULL */
-			args[i] = "rsync"; /* set a new arg0 */
-			if (verbose > 1)
-				print_child_argv("protected args:", args + i + 1);
-			fd = *f_out_p;
-			do {
-#ifdef ICONV_OPTION
-				if (convert) {
-					INIT_XBUF_STRLEN(inbuf, args[i]);
-					iconvbufs(ic_send, &inbuf, &outbuf,
-						  ICB_EXPAND_OUT | ICB_INCLUDE_BAD | ICB_INCLUDE_INCOMPLETE);
-					outbuf.buf[outbuf.len] = '\0';
-					write_buf(fd, outbuf.buf, outbuf.len + 1);
-					outbuf.len = 0;
-				} else
-#endif
-					write_buf(fd, args[i], strlen(args[i]) + 1);
-			} while (args[++i]);
-			write_byte(fd, 0);
-#ifdef ICONV_OPTION
-			if (convert)
-				free(outbuf.buf);
-#endif
-		} else
-			ret = piped_child(args, f_in_p, f_out_p);
+		if (protect_args)
+			send_protected_args(*f_out_p, args);
 	}
 
-	return ret;
+	return pid;
 
   oom:
 	out_of_memory("do_cmd");
@@ -544,12 +507,20 @@ static char *get_local_name(struct file_list *flist, char *dest_path)
 	if (!dest_path || list_only)
 		return NULL;
 
+	if (daemon_filter_list.head
+	 && (check_filter(&daemon_filter_list, FLOG, dest_path, 0 != 0) < 0
+	  || check_filter(&daemon_filter_list, FLOG, dest_path, 1 != 0) < 0)) {
+		rprintf(FERROR, "skipping daemon-excluded destination \"%s\"\n",
+			dest_path);
+		exit_cleanup(RERR_FILESELECT);
+	}
+
 	/* See what currently exists at the destination. */
 	if ((statret = do_stat(dest_path, &st)) == 0) {
 		/* If the destination is a dir, enter it and use mode 1. */
 		if (S_ISDIR(st.st_mode)) {
-			if (!push_dir(dest_path, 0)) {
-				rsyserr(FERROR, errno, "push_dir#1 %s failed",
+			if (!change_dir(dest_path, CD_NORMAL)) {
+				rsyserr(FERROR, errno, "change_dir#1 %s failed",
 					full_fname(dest_path));
 				exit_cleanup(RERR_FILESELECT);
 			}
@@ -608,8 +579,8 @@ static char *get_local_name(struct file_list *flist, char *dest_path)
 			dry_run++;
 		}
 
-		if (!push_dir(dest_path, dry_run > 1)) {
-			rsyserr(FERROR, errno, "push_dir#2 %s failed",
+		if (!change_dir(dest_path, dry_run > 1 ? CD_SKIP_CHDIR : CD_NORMAL)) {
+			rsyserr(FERROR, errno, "change_dir#2 %s failed",
 				full_fname(dest_path));
 			exit_cleanup(RERR_FILESELECT);
 		}
@@ -628,8 +599,8 @@ static char *get_local_name(struct file_list *flist, char *dest_path)
 		dest_path = "/";
 
 	*cp = '\0';
-	if (!push_dir(dest_path, 0)) {
-		rsyserr(FERROR, errno, "push_dir#3 %s failed",
+	if (!change_dir(dest_path, CD_NORMAL)) {
+		rsyserr(FERROR, errno, "change_dir#3 %s failed",
 			full_fname(dest_path));
 		exit_cleanup(RERR_FILESELECT);
 	}
@@ -721,8 +692,8 @@ static void do_server_sender(int f_in, int f_out, int argc, char *argv[])
 	}
 
 	if (!relative_paths) {
-		if (!push_dir(dir, 0)) {
-			rsyserr(FERROR, errno, "push_dir#3 %s failed",
+		if (!change_dir(dir, CD_NORMAL)) {
+			rsyserr(FERROR, errno, "change_dir#3 %s failed",
 				full_fname(dir));
 			exit_cleanup(RERR_FILESELECT);
 		}
@@ -891,8 +862,8 @@ static void do_server_recv(int f_in, int f_out, int argc, char *argv[])
 		char *dir = argv[0];
 		argc--;
 		argv++;
-		if (!am_daemon && !push_dir(dir, 0)) {
-			rsyserr(FERROR, errno, "push_dir#4 %s failed",
+		if (!am_daemon && !change_dir(dir, CD_NORMAL)) {
+			rsyserr(FERROR, errno, "change_dir#4 %s failed",
 				full_fname(dir));
 			exit_cleanup(RERR_FILESELECT);
 		}
@@ -931,25 +902,25 @@ static void do_server_recv(int f_in, int f_out, int argc, char *argv[])
 	if (sanitize_paths) {
 		char **dir_p;
 		for (dir_p = basis_dir; *dir_p; dir_p++)
-			*dir_p = sanitize_path(NULL, *dir_p, NULL, curr_dir_depth);
+			*dir_p = sanitize_path(NULL, *dir_p, NULL, curr_dir_depth, SP_DEFAULT);
 		if (partial_dir)
-			partial_dir = sanitize_path(NULL, partial_dir, NULL, curr_dir_depth);
+			partial_dir = sanitize_path(NULL, partial_dir, NULL, curr_dir_depth, SP_DEFAULT);
 	}
 	check_alt_basis_dirs();
 
-	if (server_filter_list.head) {
+	if (daemon_filter_list.head) {
 		char **dir_p;
-		struct filter_list_struct *elp = &server_filter_list;
+		struct filter_list_struct *elp = &daemon_filter_list;
 
 		for (dir_p = basis_dir; *dir_p; dir_p++) {
 			char *dir = *dir_p;
 			if (*dir == '/')
 				dir += module_dirlen;
-			if (check_filter(elp, dir, 1) < 0)
+			if (check_filter(elp, FLOG, dir, 1) < 0)
 				goto options_rejected;
 		}
 		if (partial_dir && *partial_dir == '/'
-		 && check_filter(elp, partial_dir + module_dirlen, 1) < 0) {
+		 && check_filter(elp, FLOG, partial_dir + module_dirlen, 1) < 0) {
 		    options_rejected:
 			rprintf(FERROR,
 				"Your options have been rejected by the server.\n");
@@ -1067,11 +1038,6 @@ int client_run(int f_in, int f_out, pid_t pid, int argc, char *argv[])
 			io_start_multiplex_out();
 	}
 
-	if (argc == 0) {
-		list_only |= 1;
-		xfer_dirs |= 1;
-	}
-
 	send_filter_list(read_batch ? -1 : f_out);
 
 	if (filesfrom_fd >= 0) {
@@ -1182,8 +1148,6 @@ static int start_client(int argc, char *argv[])
 				static char *dotarg[1] = { "." };
 				p = dotarg[0];
 				remote_argv = dotarg;
-				list_only |= 1;
-				xfer_dirs |= 1;
 			}
 			remote_argc = 1;
 
@@ -1255,10 +1219,6 @@ static int start_client(int argc, char *argv[])
 			}
 			remote_argv[i] = arg;
 		}
-		if (argc == 0) {
-			list_only |= 1;
-			xfer_dirs |= 1;
-		}
 	}
 
 	if (daemon_over_rsh < 0)
@@ -1447,7 +1407,7 @@ int main(int argc,char *argv[])
 	setlocale(LC_CTYPE, "");
 #endif
 
-	if (!parse_arguments(&argc, (const char ***) &argv, 1)) {
+	if (!parse_arguments(&argc, (const char ***) &argv)) {
 		/* FIXME: We ought to call the same error-handling
 		 * code here, rather than relying on getopt. */
 		option_error();
@@ -1468,11 +1428,11 @@ int main(int argc,char *argv[])
 	SIGACTION(SIGXFSZ, SIG_IGN);
 #endif
 
-	/* Initialize push_dir here because on some old systems getcwd
+	/* Initialize change_dir() here because on some old systems getcwd
 	 * (implemented by forking "pwd" and reading its output) doesn't
 	 * work when there are other child processes.  Also, on all systems
 	 * that implement getcwd that way "pwd" can't be found after chroot. */
-	push_dir(NULL, 0);
+	change_dir(NULL, CD_NORMAL);
 
 	init_flist();
 
@@ -1496,7 +1456,6 @@ int main(int argc,char *argv[])
 			read_stream_flags(batch_fd);
 		else
 			write_stream_flags(batch_fd);
-
 	}
 	if (write_batch < 0)
 		dry_run = 1;
@@ -1512,7 +1471,7 @@ int main(int argc,char *argv[])
 		char buf[MAXPATHLEN];
 		protect_args = 2;
 		read_args(STDIN_FILENO, NULL, buf, sizeof buf, 1, &argv, &argc, NULL);
-		if (!parse_arguments(&argc, (const char ***) &argv, 1)) {
+		if (!parse_arguments(&argc, (const char ***) &argv)) {
 			option_error();
 			exit_cleanup(RERR_SYNTAX);
 		}
