@@ -3,11 +3,11 @@
  *
  * Copyright (C) 1996 Andrew Tridgell
  * Copyright (C) 1996 Paul Mackerras
- * Copyright (C) 2003, 2004, 2005, 2006 Wayne Davison
+ * Copyright (C) 2003-2008 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -16,8 +16,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #include "rsync.h"
@@ -40,29 +39,50 @@ static int total_matches;
 
 extern struct stats stats;
 
-#define TABLESIZE (1<<16)
+#define TRADITIONAL_TABLESIZE (1<<16)
 
+static uint32 tablesize;
 static int32 *hash_table;
 
 #define SUM2HASH2(s1,s2) (((s1) + (s2)) & 0xFFFF)
 #define SUM2HASH(sum) SUM2HASH2((sum)&0xFFFF,(sum)>>16)
 
+#define BIG_SUM2HASH(sum) ((sum)%tablesize)
+
 static void build_hash_table(struct sum_struct *s)
 {
+	static uint32 alloc_size;
 	int32 i;
 
-	if (!hash_table) {
-		hash_table = new_array(int32, TABLESIZE);
+	/* Dynamically calculate the hash table size so that the hash load
+	 * for big files is about 80%.  A number greater than the traditional
+	 * size must be odd or s2 will not be able to span the entire set. */
+	tablesize = (uint32)(s->count/8) * 10 + 11;
+	if (tablesize < TRADITIONAL_TABLESIZE)
+		tablesize = TRADITIONAL_TABLESIZE;
+	if (tablesize > alloc_size || tablesize < alloc_size - 16*1024) {
+		if (hash_table)
+			free(hash_table);
+		hash_table = new_array(int32, tablesize);
 		if (!hash_table)
 			out_of_memory("build_hash_table");
+		alloc_size = tablesize;
 	}
 
-	memset(hash_table, 0xFF, TABLESIZE * sizeof hash_table[0]);
+	memset(hash_table, 0xFF, tablesize * sizeof hash_table[0]);
 
-	for (i = 0; i < s->count; i++) {
-		uint32 t = SUM2HASH(s->sums[i].sum1);
-		s->sums[i].chain = hash_table[t];
-		hash_table[t] = i;
+	if (tablesize == TRADITIONAL_TABLESIZE) {
+		for (i = 0; i < s->count; i++) {
+			uint32 t = SUM2HASH(s->sums[i].sum1);
+			s->sums[i].chain = hash_table[t];
+			hash_table[t] = i;
+		}
+	} else {
+		for (i = 0; i < s->count; i++) {
+			uint32 t = BIG_SUM2HASH(s->sums[i].sum1);
+			s->sums[i].chain = hash_table[t];
+			hash_table[t] = i;
+		}
 	}
 }
 
@@ -85,7 +105,7 @@ static OFF_T last_match;
 static void matched(int f, struct sum_struct *s, struct map_struct *buf,
 		    OFF_T offset, int32 i)
 {
-	int32 n = offset - last_match; /* max value: block_size (int32) */
+	int32 n = (int32)(offset - last_match); /* max value: block_size (int32) */
 	int32 j;
 
 	if (verbose > 2 && i >= 0) {
@@ -121,8 +141,8 @@ static void matched(int f, struct sum_struct *s, struct map_struct *buf,
 static void hash_search(int f,struct sum_struct *s,
 			struct map_struct *buf, OFF_T len)
 {
-	OFF_T offset, end, backup;
-	int32 k, want_i;
+	OFF_T offset, end;
+	int32 k, want_i, backup;
 	char sum2[SUM_LENGTH];
 	uint32 s1, s2, sum;
 	int more;
@@ -165,11 +185,16 @@ static void hash_search(int f,struct sum_struct *s,
 				(double)offset, s2 & 0xFFFF, s1 & 0xFFFF);
 		}
 
-		i = hash_table[SUM2HASH2(s1,s2)];
-		if (i < 0)
-			goto null_hash;
+		if (tablesize == TRADITIONAL_TABLESIZE) {
+			if ((i = hash_table[SUM2HASH2(s1,s2)]) < 0)
+				goto null_hash;
+			sum = (s1 & 0xffff) | (s2 << 16);
+		} else {
+			sum = (s1 & 0xffff) | (s2 << 16);
+			if ((i = hash_table[BIG_SUM2HASH(sum)]) < 0)
+				goto null_hash;
+		}
 
-		sum = (s1 & 0xffff) | (s2 << 16);
 		hash_hits++;
 		do {
 			int32 l;
@@ -254,7 +279,7 @@ static void hash_search(int f,struct sum_struct *s,
 		} while ((i = s->sums[i].chain) >= 0);
 
 	  null_hash:
-		backup = offset - last_match;
+		backup = (int32)(offset - last_match);
 		/* We sometimes read 1 byte prior to last_match... */
 		if (backup < 0)
 			backup = 0;
@@ -304,7 +329,8 @@ static void hash_search(int f,struct sum_struct *s,
  **/
 void match_sums(int f, struct sum_struct *s, struct map_struct *buf, OFF_T len)
 {
-	char file_sum[MD4_SUM_LENGTH];
+	char file_sum[MAX_DIGEST_LEN];
+	int sum_len;
 
 	last_match = 0;
 	false_alarms = 0;
@@ -314,22 +340,24 @@ void match_sums(int f, struct sum_struct *s, struct map_struct *buf, OFF_T len)
 
 	sum_init(checksum_seed);
 
-	if (append_mode) {
-		OFF_T j = 0;
-		for (j = CHUNK_SIZE; j < s->flength; j += CHUNK_SIZE) {
-			if (buf && do_progress)
-				show_progress(last_match, buf->file_size);
-			sum_update(map_ptr(buf, last_match, CHUNK_SIZE),
-				   CHUNK_SIZE);
-			last_match = j;
-		}
-		if (last_match < s->flength) {
-			int32 len = s->flength - last_match;
-			if (buf && do_progress)
-				show_progress(last_match, buf->file_size);
-			sum_update(map_ptr(buf, last_match, len), len);
-			last_match = s->flength;
+	if (append_mode > 0) {
+		if (append_mode == 2) {
+			OFF_T j = 0;
+			for (j = CHUNK_SIZE; j < s->flength; j += CHUNK_SIZE) {
+				if (buf && do_progress)
+					show_progress(last_match, buf->file_size);
+				sum_update(map_ptr(buf, last_match, CHUNK_SIZE),
+					   CHUNK_SIZE);
+				last_match = j;
+			}
+			if (last_match < s->flength) {
+				int32 n = (int32)(s->flength - last_match);
+				if (buf && do_progress)
+					show_progress(last_match, buf->file_size);
+				sum_update(map_ptr(buf, last_match, n), n);
+			}
 		}
+		last_match = s->flength;
 		s->count = 0;
 	}
 
@@ -339,7 +367,7 @@ void match_sums(int f, struct sum_struct *s, struct map_struct *buf, OFF_T len)
 		if (verbose > 2)
 			rprintf(FINFO,"built hash table\n");
 
-		hash_search(f,s,buf,len);
+		hash_search(f, s, buf, len);
 
 		if (verbose > 2)
 			rprintf(FINFO,"done hash search\n");
@@ -351,14 +379,14 @@ void match_sums(int f, struct sum_struct *s, struct map_struct *buf, OFF_T len)
 		matched(f, s, buf, len, -1);
 	}
 
-	sum_end(file_sum);
+	sum_len = sum_end(file_sum);
 	/* If we had a read error, send a bad checksum. */
 	if (buf && buf->status != 0)
 		file_sum[0]++;
 
 	if (verbose > 2)
 		rprintf(FINFO,"sending file_sum\n");
-	write_buf(f,file_sum,MD4_SUM_LENGTH);
+	write_buf(f, file_sum, sum_len);
 
 	if (verbose > 2)
 		rprintf(FINFO, "false_alarms=%d hash_hits=%d matches=%d\n",
