@@ -4,11 +4,11 @@
  * Copyright (C) 1996-2000 Andrew Tridgell
  * Copyright (C) 1996 Paul Mackerras
  * Copyright (C) 2001, 2002 Martin Pool <mbp@samba.org>
- * Copyright (C) 2003, 2004, 2005, 2006 Wayne Davison
+ * Copyright (C) 2003-2008 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -17,14 +17,15 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #include "rsync.h"
 
-extern struct stats stats;
 extern int am_server;
+extern int need_unsorted_flist;
+extern struct stats stats;
+extern struct file_list *cur_flist;
 
 #define PROGRESS_HISTORY_SECS 5
 
@@ -42,6 +43,7 @@ struct progress_history {
 static struct progress_history ph_start;
 static struct progress_history ph_list[PROGRESS_HISTORY_SECS];
 static int newest_hpos, oldest_hpos;
+static int current_file_index;
 
 static unsigned long msdiff(struct timeval *t1, struct timeval *t2)
 {
@@ -60,12 +62,11 @@ static unsigned long msdiff(struct timeval *t1, struct timeval *t2)
 static void rprint_progress(OFF_T ofs, OFF_T size, struct timeval *now,
 			    int is_last)
 {
-	char eol[256];
+	char rembuf[64], eol[128];
 	const char *units;
 	int pct = ofs == size ? 100 : (int) (100.0 * ofs / size);
 	unsigned long diff;
 	double rate, remain;
-	int remain_h, remain_m, remain_s;
 
 	if (is_last) {
 		/* Compute stats based on the starting info. */
@@ -94,20 +95,33 @@ static void rprint_progress(OFF_T ofs, OFF_T size, struct timeval *now,
 		units = "kB/s";
 	}
 
-	remain_s = (int) remain % 60;
-	remain_m = (int) (remain / 60.0) % 60;
-	remain_h = (int) (remain / 3600.0);
+	if (remain < 0)
+		strlcpy(rembuf, "  ??:??:??", sizeof rembuf);
+	else {
+		snprintf(rembuf, sizeof rembuf, "%4d:%02d:%02d",
+			 (int) (remain / 3600.0),
+			 (int) (remain / 60.0) % 60,
+			 (int) remain % 60);
+	}
 
 	if (is_last) {
 		snprintf(eol, sizeof eol, " (xfer#%d, to-check=%d/%d)\n",
 			stats.num_transferred_files,
-			stats.num_files - stats.current_file_index - 1,
+			stats.num_files - current_file_index - 1,
 			stats.num_files);
 	} else
 		strlcpy(eol, "\r", sizeof eol);
-	rprintf(FCLIENT, "%12s %3d%% %7.2f%s %4d:%02d:%02d%s",
-		human_num(ofs), pct, rate, units,
-		remain_h, remain_m, remain_s, eol);
+	rprintf(FCLIENT, "%12s %3d%% %7.2f%s %s%s",
+		human_num(ofs), pct, rate, units, rembuf, eol);
+}
+
+void set_current_file_index(struct file_struct *file, int ndx)
+{
+	if (need_unsorted_flist)
+		current_file_index = flist_find(cur_flist, file) + cur_flist->ndx_start;
+	else
+		current_file_index = ndx;
+	current_file_index -= cur_flist->flist_num;
 }
 
 void end_progress(OFF_T size)
