@@ -49,8 +49,8 @@
 
 # This script must be invoked from the build directory.  
 
-# A scratch directory, 'testtmp', is created in the build directory to
-# hold working files.
+# A scratch directory, 'testtmp', is used in the build directory to
+# hold per-test subdirectories.
 
 # This script also uses the $loglevel environment variable.  1 is the
 # default value, and 10 the most verbose.  You can set this from the
@@ -181,12 +181,15 @@ else
     echo "    preserve_scratch=no"
 fi    
 
-# Check if setfacl is around and if it supports the -k or -s option.
-if setfacl --help 2>&1 | grep ' -k,\|\[-[a-z]*k' >/dev/null; then
+# Check if setacl/setfacl is around and if it supports the -k or -s option.
+if setacl -k u::7,g::5,o:5 testsuite 2>/dev/null; then
+    setfacl_nodef='setacl -k'
+elif setfacl --help 2>&1 | grep ' -k,\|\[-[a-z]*k' >/dev/null; then
     setfacl_nodef='setfacl -k'
 elif setfacl -s u::7,g::5,o:5 testsuite 2>/dev/null; then
     setfacl_nodef='setfacl -s u::7,g::5,o:5'
 else
+    # The "true" command runs successfully, but does nothing.
     setfacl_nodef=true
 fi
 
@@ -207,22 +210,26 @@ missing=0
 passed=0
 failed=0
 
-# Prefix for scratch directory.  We create separate directories for
-# each test case, so that they can be left behind in case of failure
-# to aid investigation.
+# Directory that holds the other test subdirs.  We create separate dirs
+# inside for each test case, so that they can be left behind in case of
+# failure to aid investigation.  We don't remove the testtmp subdir at
+# the end so that it can be configured as a symlink to a filesystem that
+# has ACLs and xattr support enabled (if desired).
 scratchbase="$TOOLDIR"/testtmp
 echo "    scratchbase=$scratchbase"
+[ -d "$scratchbase" ] || mkdir "$scratchbase"
 
 suitedir="$srcdir/testsuite"
 
 export scratchdir suitedir
 
 prep_scratch() {
-    [ -d "$scratchdir" ] && rm -rf "$scratchdir"
+    [ -d "$scratchdir" ] && chmod -R u+rwX "$scratchdir" && rm -rf "$scratchdir"
     mkdir "$scratchdir"
     # Get rid of default ACLs and dir-setgid to avoid confusing some tests.
     $setfacl_nodef "$scratchdir" || true
     chmod g-s "$scratchdir"
+    ln -s "$srcdir" "$scratchdir/src"
     return 0
 }
 
@@ -238,7 +245,7 @@ fi
 for testscript in $suitedir/$whichtests
 do
     testbase=`echo $testscript | sed -e 's!.*/!!' -e 's/.test\$//'`
-    scratchdir="$scratchbase.$testbase"
+    scratchdir="$scratchbase/$testbase"
 
     prep_scratch
 
